@@ -5,32 +5,20 @@ proc log_file_matches {log pattern} {
     string match $pattern $content
 }
 
-proc get_client_id_by_last_cmd {r cmd} {
-    set client_list [$r client list]
-    set client_id ""
-    set lines [split $client_list "\n"]
-    foreach line $lines {
-        if {[string match *cmd=$cmd* $line]} {
-            set parts [split $line " "]
-            foreach part $parts {
-                if {[string match id=* $part]} {
-                    set client_id [lindex [split $part "="] 1]
-                    return $client_id
-                }
-            }
-        }
-    }
-    return $client_id
-}
-
-# Wait until the process enters a paused state, then resume the process.
-proc wait_and_resume_process idx {
+# Wait until the process enters a paused state.
+proc wait_process_paused idx {
     set pid [srv $idx pid]
     wait_for_condition 50 1000 {
         [string match "T*" [exec ps -o state= -p $pid]]
     } else {
         fail "Process $pid didn't stop, current state is [exec ps -o state= -p $pid]"
     }
+}
+
+# Wait until the process enters a paused state, then resume the process.
+proc wait_and_resume_process idx {
+    set pid [srv $idx pid]
+    wait_process_paused $idx
     resume_process $pid
 }
 
@@ -146,6 +134,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
 
         $primary config set rdb-key-save-delay 200
         $primary config set dual-channel-replication-enabled yes
+        $primary config set repl-diskless-sync-delay 0
         $replica config set dual-channel-replication-enabled yes
         $replica config set repl-diskless-sync no
 
@@ -237,6 +226,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
             # a replication buffer block.
             $primary config set client-output-buffer-limit "replica 1100k 0 0"
             $primary config set dual-channel-replication-enabled $enable
+            $primary config set repl-diskless-sync-delay 0
             $replica config set dual-channel-replication-enabled $enable
 
             test "Toggle dual-channel-replication-enabled: $enable start" {    
@@ -521,7 +511,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
             }
             wait_for_value_to_propagate_to_replica $primary $replica "key1"
             # Confirm the occurrence of a race condition.
-            wait_for_log_messages -1 {"*Dual channel sync - psync established after rdb load*"} 0 2000 1
+            wait_for_log_messages -1 {"*Dual channel replication: Psync established after rdb load*"} 0 2000 1
         }
     }
 }
@@ -542,7 +532,14 @@ start_server {tags {"dual-channel-replication external:skip"}} {
         $primary config set dual-channel-replication-enabled yes
         $primary config set repl-backlog-size $backlog_size
         $primary config set loglevel debug
-        $primary config set repl-timeout 10
+        $primary config set repl-diskless-sync-delay 0
+        if {$::valgrind} {
+            $primary config set repl-timeout 100
+            $replica config set repl-timeout 100
+        } else {
+            $primary config set repl-timeout 10            
+            $replica config set repl-timeout 10
+        }
         $primary config set rdb-key-save-delay 200
         populate 10000 primary 10000
         
@@ -552,7 +549,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
 
         $replica config set dual-channel-replication-enabled yes
         $replica config set loglevel debug
-        $replica config set repl-timeout 10
+        
         # Pause replica after primary fork
         $replica debug pause-after-fork 1
 
@@ -800,10 +797,11 @@ start_server {tags {"dual-channel-replication external:skip"}} {
         set load_handle0 [start_write_load $primary_host $primary_port 60]
         set load_handle1 [start_write_load $primary_host $primary_port 60]
         set load_handle2 [start_write_load $primary_host $primary_port 60]
+        set replica_loglines [count_log_lines 0]
 
         $replica config set dual-channel-replication-enabled yes
         $replica config set loglevel debug
-        $replica config set repl-timeout 10
+        $replica config set repl-timeout 60
         $primary config set repl-backlog-size 1mb
 
         test "Test dual-channel-replication primary gets cob overrun before established psync" {
@@ -824,18 +822,59 @@ start_server {tags {"dual-channel-replication external:skip"}} {
             } else {
                 fail "Primary did not free repl buf block after sync failure"
             }
+            # Full sync will be triggered after the replica is reconnected, pause primary main process after fork.
+            # In this way, in the subsequent replicaof no one, we won't get the LOADING error if the replica reconnects
+            # too quickly and enters the loading state.
+            $primary debug pause-after-fork 1
+            set replica_loglines [count_log_lines 0]
             resume_process $replica_pid
             set res [wait_for_log_messages -1 {"*Unable to partial resync with replica * for lack of backlog*"} $loglines 2000 10]
             set loglines [lindex $res 1]
         }
+        # Waiting for the primary to enter the paused state, that is, make sure that bgsave is triggered.
+        wait_process_paused -1
+        wait_for_log_messages 0 {"*Done loading RDB*"} $replica_loglines 1000 10
         $replica replicaof no one
+        # Resume the primary and make sure the sync is dropped.
+        resume_process [srv -1 pid]
+        $primary debug pause-after-fork 0
         wait_for_condition 500 1000 {
             [s -1 rdb_bgsave_in_progress] eq 0
         } else {
             fail "Primary should abort sync"
         }
+        stop_write_load $load_handle0
+        stop_write_load $load_handle1
+        stop_write_load $load_handle2
+    }
+}
+
+start_server {tags {"dual-channel-replication external:skip"}} {
+    set primary [srv 0 client]
+    set primary_host [srv 0 host]
+    set primary_port [srv 0 port]
+    set loglines [count_log_lines 0]
+
+    $primary config set repl-diskless-sync yes
+    $primary config set dual-channel-replication-enabled yes
+    $primary config set client-output-buffer-limit "replica 1100k 0 0"
+    $primary config set loglevel debug
+    start_server {} {
+        set replica [srv 0 client]
+        set replica_host [srv 0 host]
+        set replica_port [srv 0 port]
+        set replica_log [srv 0 stdout]
+        set replica_pid  [srv 0 pid]
+        
+        set load_handle0 [start_write_load $primary_host $primary_port 60]
+        set load_handle1 [start_write_load $primary_host $primary_port 60]
+        set load_handle2 [start_write_load $primary_host $primary_port 60]
+
+        $replica config set dual-channel-replication-enabled yes
+        $replica config set loglevel debug
+        $replica config set repl-timeout 60
+        $primary config set repl-backlog-size 1mb
         
-        $replica debug pause-after-fork 1
         $primary debug populate 1000 primary 100000
         # Set primary with a slow rdb generation, so that we can easily intercept loading
         # 10ms per key, with 1000 keys is 10 seconds
@@ -843,6 +882,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
 
         test "Test dual-channel-replication primary gets cob overrun during replica rdb load" {
             set cur_client_closed_count [s -1 client_output_buffer_limit_disconnections]
+            $replica debug pause-after-fork 1
             $replica replicaof $primary_host $primary_port
             wait_for_condition 500 1000 {
                 [s -1 client_output_buffer_limit_disconnections] > $cur_client_closed_count
@@ -855,7 +895,12 @@ start_server {tags {"dual-channel-replication external:skip"}} {
             } else {
                 fail "Primary did not free repl buf block after sync failure"
             }
+
+            # Increase the delay to make sure the replica doesn't start another sync
+            # after it resumes after the first one.
+            $primary config set repl-diskless-sync-delay 100
             wait_and_resume_process 0
+            $replica debug pause-after-fork 0
             set res [wait_for_log_messages -1 {"*Unable to partial resync with replica * for lack of backlog*"} $loglines 20000 1]
             set loglines [lindex $res 0]
         }
@@ -865,7 +910,6 @@ start_server {tags {"dual-channel-replication external:skip"}} {
     }
 }
 
-foreach dualchannel {yes no} {
 start_server {tags {"dual-channel-replication external:skip"}} {
     set primary [srv 0 client]
     set primary_host [srv 0 host]
@@ -876,25 +920,25 @@ start_server {tags {"dual-channel-replication external:skip"}} {
     $primary config set dual-channel-replication-enabled yes
     $primary config set loglevel debug
     $primary config set repl-diskless-sync-delay 5
-    
+    $primary config set client-output-buffer-limit "replica 0 0 0"
+
     # Generating RDB will cost 5s(10000 * 0.0005s)
     $primary debug populate 10000 primary 1
     $primary config set rdb-key-save-delay 500
-
-    $primary config set dual-channel-replication-enabled $dualchannel
+    $primary config set dual-channel-replication-enabled yes
 
     start_server {} {
         set replica1 [srv 0 client]
-        $replica1 config set dual-channel-replication-enabled $dualchannel
+        $replica1 config set dual-channel-replication-enabled yes
         $replica1 config set loglevel debug
         start_server {} {
             set replica2 [srv 0 client]
-            $replica2 config set dual-channel-replication-enabled $dualchannel
+            $replica2 config set dual-channel-replication-enabled yes
             $replica2 config set loglevel debug
             $replica2 config set repl-timeout 60
 
             set load_handle [start_one_key_write_load $primary_host $primary_port 100 "mykey1"]
-            test "Sync should continue if not all slaves dropped dual-channel-replication $dualchannel" {
+            test "Sync should continue if not all slaves dropped" {
                 $replica1 replicaof $primary_host $primary_port
                 $replica2 replicaof $primary_host $primary_port
 
@@ -903,22 +947,19 @@ start_server {tags {"dual-channel-replication external:skip"}} {
                 } else {
                     fail "Sync did not start"
                 }
-                if {$dualchannel == "yes"} {
-                    # Wait for both replicas main conns to establish psync
-                    wait_for_condition 50 1000 {
-                        [status $primary sync_partial_ok] == 2
-                    } else {
-                        fail "Replicas main conns didn't establish psync [status $primary sync_partial_ok]"
-                    }
+                # Wait for both replicas main conns to establish psync
+                wait_for_condition 50 1000 {
+                    [status $primary sync_partial_ok] == 2
+                } else {
+                    fail "Replicas main conns didn't establish psync [status $primary sync_partial_ok]"
                 }
-
                 catch {$replica1 shutdown nosave}
                 wait_for_condition 50 2000 {
                     [status $replica2 master_link_status] == "up" &&
                     [status $primary sync_full] == 2 &&
-                    (($dualchannel == "yes" && [status $primary sync_partial_ok] == 2) || $dualchannel == "no")
+                    ([status $primary sync_partial_ok] == 2)
                 } else {
-                    fail "Sync session interapted\n
+                    fail "Sync session interrupted\n
                         sync_full:[status $primary sync_full]\n
                         sync_partial_ok:[status $primary sync_partial_ok]"
                 }
@@ -930,7 +971,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
             $primary debug populate 1000000 primary 1
             $primary config set rdb-key-save-delay 100
     
-            test "Primary abort sync if all slaves dropped dual-channel-replication $dualchannel" {
+            test "Primary abort sync if all slaves dropped dual-channel-replication" {
                 set cur_psync [status $primary sync_partial_ok]
                 $replica2 replicaof $primary_host $primary_port
 
@@ -939,13 +980,11 @@ start_server {tags {"dual-channel-replication external:skip"}} {
                 } else {
                     fail "Sync did not start"
                 }
-                if {$dualchannel == "yes"} {
-                    # Wait for both replicas main conns to establish psync
-                    wait_for_condition 50 1000 {
-                        [status $primary sync_partial_ok] == $cur_psync + 1
-                    } else {
-                        fail "Replicas main conns didn't establish psync [status $primary sync_partial_ok]"
-                    }
+                # Wait for both replicas main conns to establish psync
+                wait_for_condition 50 1000 {
+                    [status $primary sync_partial_ok] == $cur_psync + 1
+                } else {
+                    fail "Replicas main conns didn't establish psync [status $primary sync_partial_ok]"
                 }
 
                 catch {$replica2 shutdown nosave}
@@ -959,7 +998,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
         }
     }
 }
-}
+
 
 start_server {tags {"dual-channel-replication external:skip"}} {
     set primary [srv 0 client]
@@ -970,8 +1009,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
     $primary config set repl-diskless-sync yes
     $primary config set dual-channel-replication-enabled yes
     $primary config set loglevel debug
-    $primary config set repl-diskless-sync-delay 5; # allow catch failed sync before retry
-
+    $primary config set repl-diskless-sync-delay 0
     # Generating RDB will cost 500s(1000000 * 0.0001s)
     $primary debug populate 1000000 primary 1
     $primary config set rdb-key-save-delay 100
@@ -1002,6 +1040,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
             set replica_main_conn_id [get_client_id_by_last_cmd $primary "psync"]
             assert {$replica_main_conn_id != ""}
             set loglines [count_log_lines -1]
+            $primary config set repl-diskless-sync-delay 5; # allow catch failed sync before retry
             $primary client kill id $replica_main_conn_id
             # Wait for primary to abort the sync
             wait_for_condition 50 1000 {
@@ -1022,6 +1061,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
         }
 
         test "Test dual-channel-replication replica rdb connection disconnected" {
+            $primary config set repl-diskless-sync-delay 0
             $replica replicaof $primary_host $primary_port
             # Wait for sync session to start
             wait_for_condition 500 1000 {
@@ -1036,6 +1076,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
             $primary debug log "killing replica rdb connection $replica_rdb_channel_id"
             assert {$replica_rdb_channel_id != ""}
             set loglines [count_log_lines -1]
+            $primary config set repl-diskless-sync-delay 5; # allow catch failed sync before retry
             $primary client kill id $replica_rdb_channel_id
             # Wait for primary to abort the sync
             wait_for_log_messages -1 {"*Background RDB transfer error*"} $loglines 1000 10
@@ -1051,6 +1092,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
         }
 
         test "Test dual-channel-replication primary reject set-rdb-client after client killed" {
+            $primary config set repl-diskless-sync-delay 0
             # Ensure replica main channel will not handshake before rdb client is killed
             $replica debug pause-after-fork 1
             $replica replicaof $primary_host $primary_port
@@ -1065,6 +1107,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
             set replica_rdb_channel_id [get_client_id_by_last_cmd $primary "sync"]
             assert {$replica_rdb_channel_id != ""}
             $primary debug log "killing replica rdb connection $replica_rdb_channel_id"
+            $primary config set repl-diskless-sync-delay 5; # allow catch failed sync before retry
             $primary client kill id $replica_rdb_channel_id
             # Wait for primary to abort the sync
             wait_and_resume_process 0
@@ -1142,7 +1185,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
     $primary config set repl-diskless-sync yes
     $primary config set dual-channel-replication-enabled yes
     $primary config set loglevel debug
-    $primary config set repl-diskless-sync-delay 5; # allow catch failed sync before retry
+    $primary config set repl-diskless-sync-delay 0
 
     # Generating RDB will cost 100 sec to generate
     $primary debug populate 100000 primary 1
@@ -1157,6 +1200,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
         $replica config set dual-channel-replication-enabled yes
         $replica config set loglevel debug
         $replica config set repl-timeout 10
+        $replica config set loading-process-events-interval-bytes 1024
         set load_handle [start_one_key_write_load $primary_host $primary_port 100 "mykey"]
         test "Replica recover rdb-connection killed" {
             $replica replicaof $primary_host $primary_port
@@ -1173,6 +1217,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
             set replica_rdb_channel_id [get_client_id_by_last_cmd $primary "sync"]
             assert {$replica_rdb_channel_id != ""}
             set loglines [count_log_lines -1]
+            $primary config set repl-diskless-sync-delay 5; # allow catch failed sync before retry
             $primary client kill id $replica_rdb_channel_id
             # Wait for primary to abort the sync
             wait_for_condition 50 1000 {
@@ -1180,6 +1225,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
             } else {
                 fail "Primary did not free repl buf block after sync failure"
             }
+            $primary config set repl-diskless-sync-delay 0
             wait_for_log_messages -1 {"*Background RDB transfer error*"} $loglines 1000 10
             # Replica should retry
             wait_for_condition 500 1000 {
@@ -1188,14 +1234,38 @@ start_server {tags {"dual-channel-replication external:skip"}} {
                 [s -1 rdb_bgsave_in_progress] eq 1
             } else {
                 fail "replica didn't retry after connection close"
-            }            
-        }
-        $replica replicaof no one
-        wait_for_condition 500 1000 {
-            [s -1 rdb_bgsave_in_progress] eq 0
-        } else {
-            fail "Primary should abort sync"
+            }
         }
+        stop_write_load $load_handle
+    }
+}
+
+start_server {tags {"dual-channel-replication external:skip"}} {
+    set primary [srv 0 client]
+    set primary_host [srv 0 host]
+    set primary_port [srv 0 port]
+    set loglines [count_log_lines 0]
+
+    $primary config set repl-diskless-sync yes
+    $primary config set dual-channel-replication-enabled yes
+    $primary config set loglevel debug
+    $primary config set repl-diskless-sync-delay 0
+
+    # Generating RDB will cost 100 sec to generate
+    $primary debug populate 100000 primary 1
+    $primary config set rdb-key-save-delay 1000
+    
+    start_server {} {
+        set replica [srv 0 client]
+        set replica_host [srv 0 host]
+        set replica_port [srv 0 port]
+        set replica_log [srv 0 stdout]
+        
+        $replica config set dual-channel-replication-enabled yes
+        $replica config set loglevel debug
+        $replica config set repl-timeout 10
+        $replica config set loading-process-events-interval-bytes 1024
+        set load_handle [start_one_key_write_load $primary_host $primary_port 100 "mykey"]
         test "Replica recover main-connection killed" {
             $replica replicaof $primary_host $primary_port
             # Wait for sync session to start
@@ -1206,11 +1276,11 @@ start_server {tags {"dual-channel-replication external:skip"}} {
             } else {
                 fail "replica didn't start sync session in time"
             }            
-
             $primary debug log "killing replica main connection"
             set replica_main_conn_id [get_client_id_by_last_cmd $primary "psync"]
             assert {$replica_main_conn_id != ""}
             set loglines [count_log_lines -1]
+            $primary config set repl-diskless-sync-delay 5; # allow catch failed sync before retry
             $primary client kill id $replica_main_conn_id
             # Wait for primary to abort the sync
             wait_for_condition 50 1000 {
@@ -1218,6 +1288,7 @@ start_server {tags {"dual-channel-replication external:skip"}} {
             } else {
                 fail "Primary did not free repl buf block after sync failure"
             }
+            $primary config set repl-diskless-sync-delay 0
             wait_for_log_messages -1 {"*Background RDB transfer error*"} $loglines 1000 10
             # Replica should retry
             wait_for_condition 500 1000 {
@@ -1232,69 +1303,69 @@ start_server {tags {"dual-channel-replication external:skip"}} {
     }
 }
 
-# Test Sequence:
-# 1. replica -> primary.
-# 2. Replica initiates synchronization via RDB channel.
-# 3. Primary's main process is suspended.
-# 4. Replica completes RDB loading and pauses before establishing PSYNC connection.
-# 5. Primary resumes operation and detects closed RDB channel.
-# 6. Primary protects the RDB channel, maintains the RDB channel.
-# 7. replica -> primary -> new_primary
-# 8. Starting a new server, set up a chained replica.
-# 9. The primary completes sync from the new_primary server and disconnects all replica
-#    clients (including the RDB channel).
-# 10. Make sure that the primary does not assert.
-# 11. Replica resumes operation, check the replication is working correctly.
-test "Chained replicas can disconnect protected RDB channel client when using dual channel replication" {
-    start_server {tags {"dual-channel-replication external:skip"}} {
-        set primary [srv 0 client]
-        set primary_host [srv 0 host]
-        set primary_port [srv 0 port]
 
-        $primary config set repl-diskless-sync yes
-        $primary config set dual-channel-replication-enabled yes
-        $primary debug pause-after-fork 1
-        $primary debug delay-rdb-client-free-seconds 60
+start_server {tags {"dual-channel-replication external:skip"}} {
+    set primary [srv 0 client]
+    set primary_host [srv 0 host]
+    set primary_port [srv 0 port]
 
-        start_server {} {
-            set replica [srv 0 client]
-            set replica_pid [srv 0 pid]
+    $primary config set repl-diskless-sync yes
+    $primary config set dual-channel-replication-enabled yes
+    $primary config set repl-diskless-sync-delay 0
 
-            $replica config set dual-channel-replication-enabled yes
+    # Generating RDB will take 100 sec to generate
+    $primary debug populate 1000000 primary 1
+    $primary config set rdb-key-save-delay 10
+    
+    start_server {} {
+        set replica [srv 0 client]
+        set replica_host [srv 0 host]
+        set replica_port [srv 0 port]
+        set replica_log [srv 0 stdout]
 
-            # The primary will protect the replica's RDB channel.
+        $replica config set dual-channel-replication-enabled yes
+        $replica config set loglevel debug
+        $replica config set repl-diskless-load flush-before-load
+        $replica config set loading-process-events-interval-bytes 1024
+
+        if {$::valgrind} {
+            $primary config set repl-timeout 100
+            $replica config set repl-timeout 100
+            set max_tries 5000
+        } else {
+            $primary config set repl-timeout 10
+            $replica config set repl-timeout 10
+            set max_tries 1000
+        }
+
+        test "Replica notice main-connection killed during rdb load callback" {; # https://github.com/valkey-io/valkey/issues/1152
             set loglines [count_log_lines 0]
             $replica replicaof $primary_host $primary_port
-            wait_for_log_messages 0 {"*Done loading RDB*"} $loglines 1000 50
-            pause_process $replica_pid
-            wait_and_resume_process -1
-            $primary debug pause-after-fork 0
-            wait_for_condition 1000 50 {
-                [string match {*replicas_waiting_psync:1*} [$primary info replication]]
+            # Wait for sync session to start
+            wait_for_condition 500 1000 {
+                [string match "*slave*,state=wait_bgsave*,type=rdb-channel*" [$primary info replication]] &&
+                [string match "*slave*,state=bg_transfer*,type=main-channel*" [$primary info replication]] &&
+                [s -1 rdb_bgsave_in_progress] eq 1
             } else {
-                fail "Primary freed RDB client before psync was established"
+                fail "replica didn't start sync session in time"
             }
-
-            start_server {} {
-                set new_primary [srv 0 client]
-                set new_primary_host [srv 0 host]
-                set new_primary_port [srv 0 port]
-
-                # Doing the chained replica, make sure it won't assert.
-                set loglines [count_log_lines -2]
-                $primary replicaof $new_primary_host $new_primary_port
-                wait_for_log_messages -2 {"*Done loading RDB*"} $loglines 1000 50
-
-                # Check the replication is working correctly.
-                resume_process $replica_pid
-                $new_primary set foo bar
-                wait_for_condition 1000 50 {
-                    [$primary get foo] eq "bar" &&
-                    [$replica get foo] eq "bar"
-                } else {
-                    fail "Chained replicas did not sync"
-                }
+            wait_for_log_messages 0 {"*Loading RDB produced by Valkey version*"} $loglines 1000 10
+            $primary set key val
+            set replica_main_conn_id [get_client_id_by_last_cmd $primary "psync"]
+            $primary config set repl-diskless-sync-delay 5; # allow catch failed sync before retry
+            $primary debug log "killing replica main connection $replica_main_conn_id"
+            assert {$replica_main_conn_id != ""}
+            set loglines [count_log_lines 0]
+            $primary config set rdb-key-save-delay 0; # disable delay to allow next sync to succeed
+            $primary client kill id $replica_main_conn_id
+            # Wait for primary to abort the sync
+            wait_for_condition 50 1000 {
+                [string match {*replicas_waiting_psync:0*} [$primary info replication]]
+            } else {
+                fail "Primary did not free repl buf block after sync failure"
             }
+            wait_for_log_messages 0 {"*Failed trying to load the PRIMARY synchronization DB from socket*"} $loglines $max_tries 10
+            verify_replica_online $primary 0 $max_tries
         }
     }
 }
