@@ -47,6 +47,10 @@ start_server {tags {"keyspace"}} {
         r dbsize
     } {0}
 
+    test {KEYS with empty DB} {
+        assert_equal {} [r keys *]
+    }
+
     test "DEL against expired key" {
         r debug set-active-expire 0
         r setex keyExpire 1 valExpire
@@ -524,7 +528,7 @@ foreach {type large} [array get largevalue] {
        assert_error "ERR invalid first DB index*" {r swapdb a 55}
        assert_error "ERR invalid first DB index*" {r swapdb a b}
        assert_match "OK" [r swapdb 0 0]
-    } {} {singledb:skip}
+    } {} {singledb:skip cluster:skip}
 
     test {Coverage: SWAPDB and FLUSHDB} {
        # set a key in each db and swapdb one of 2 with different db
@@ -546,5 +550,22 @@ foreach {type large} [array get largevalue] {
        assert_no_match "*db0:keys=*" [r info keyspace]
        assert_no_match "*db2:keys=*" [r info keyspace]
        r flushall
-    } {OK} {singledb:skip}
+    } {OK} {singledb:skip cluster:skip}
+
+    test {Regression for pattern matching very long nested loops} {
+        r flushdb
+        r SET [string repeat "a" 50000] 1
+        r KEYS [string repeat "*?" 50000]
+    } {}
+}
+
+start_cluster 1 0 {tags {"keyspace external:skip cluster"}} {
+    test {KEYS with empty DB in cluster mode} {
+        assert_equal {} [r keys *]
+        assert_equal {} [r keys foo*]
+    }
+
+    test {KEYS with empty slot in cluster mode} {
+        assert_equal {} [r keys foo]
+    }
 }
